# FILE: eventHandler.tcl
#
# The eventHandler class handles events generated by em_classes and their
# instances in a manner analogous to the Tk commands bind, bindtags, and
# event.
#
# Each instance of the eventHandler class stores a binding tag, an event
# identifier, and a Tcl script, indicating that when the event named by
# the identifier occurs in an event-driven object tagged with the binding 
# tag, the script should be evaluated to handle the event.
#
# Each object which will generate events must keep a list of the binding
# tags it is associated with, typically its own name and the name of its
# class.  Say this list is stored in a variable 'bindtags'.  An object 
# generates an event via a call to 'em_eventHandler Generate <event> 
# <bindtags>', indicating that the evene <event> has occurred in the object
# and handlers for each of its binding tags should be invoked.
#
# Each object which wishes to handle events should create suitbale new 
# instances of em_eventHandler, and take care to delete them when they
# are no longer suitable.
#
# Some care may need to be taken that when the invocation of an eventHandler
# generates new events, the recursion has a 'bottom' so an infinite loop
# is not entered.
#
# Binding tags and event id's may be any non-empty string.
#
# TODO:	The binding tags associated with any instance of any em_class
# should be encapsulated within em_eventHandler instead of having each
# object manage its own list.  Copy the Tk bindtags interface.

em_class em_eventHandler {

    # A map from (tag,event) pairs to handler lists
    common map {[em_array new]}

    common log {[em_extension ApplicationLog]}

    proc Generate { event taglist } {
        foreach tag $taglist {
            $log Append "Event: $tag $event"
            if {[catch {$map lookup $tag,$event} handlers]} {
                continue
            } else {
                foreach handler $handlers {
                    # Have to check existence of handler, because eval-ing
                    # one handler in the list may have the effect of deleting
                    # a later handler in the list.
                    if {[catch {$handler Script} script]} {
                        continue
                    } else {
                        # Handlers execute at global scope -- Good idea?
                        uplevel #0 $script
                    }
                }
            }
        }
    }

    proc Bindings { tag {event ""}} {
        if {[string match {} $event]} {
            set ret {}
            foreach pair [$map names $tag,*] {
                lappend ret [string range $pair [string length $tag,] end]
            }
            return $ret
        }
        if {[catch {$map lookup $tag,$event} handlers]} {
            return ""
        }
        return $handlers
    }

    # The binding tag for this handler instance
    private variable tag

    # The event identifier for this handler instance
    private variable event

    # The script to evaluate to handle the event
    private variable script

    constructor {_tag _event _script {position end}} {
        if {[string match {} $_tag]} {
            error "The binding tag must be non-empty"
        } else {
            set tag $_tag
        }
        if {[string match {} $_event]} {
            error "The event id must be non-empty"
        } else {
            set event $_event
        }
        if {[info complete $_script]} {
            set script $_script
        } else {
            error "The handler script provided is not a complete Tcl command:
	$_script"
        }
        # Register self on appropriate list of event handlers
        if {[catch {$map lookup $tag,$event} handlers]} {
            # Initialize handlers to empty list
            set handlers {}
            $map set [list $tag,$event {}]
        }
        $map set [list $tag,$event [linsert $handlers $position $this]]
    }

    private method Script {} {
        return $script
    }

    destructor {
        # Deregister self on appropriate list of event handlers
        set handlers [$map lookup $tag,$event]
        set i [lsearch -exact $handlers $this]
        if {$i < 0} {
            error "Programming error: eventHandler $this disappeared!"
        } else {
            $map set [list $tag,$event [lreplace $handlers $i $i]]
        }
    }
  
}

