/* FILE: misc.cc                 -*-Mode: c++-*-
 *
 * Useful miscellaneous stuff.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "misc.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

int Verbosity=10; // Default value for Verbosity
#define BigBufSize   4096   // Size used for large (normally char) buffers
#define MedBufSize   1024   // "" medium sized buffers
#define SmallBufSize  256   // "" small buffers

const char* ErrArrOutOfBounds = "Array index out of bounds";
const char* ErrBadFile        = "File error";
const char* ErrBadFileHeader  = "Bad file header";
const char* ErrBadFileOpen    = "Unable to open file";
const char* ErrInputEOF       = "Premature End-Of-File on input";
const char* ErrOutputEOF      = "Premature End-Of-File on output";
const char* ErrBadParam       = "Bad parameter";
const char* ErrDispInit       = "Unable to initialize display";
const char* ErrMisc           = "Miscellaneous error";
const char* ErrNoMem          = "Insufficent working memory";
const char* ErrOverflow       = "Arithmetic overflow";
const char* ErrProgramming    = "Programming error";

ClassDoc::ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
		   const CHAR *new_revision,const CHAR *new_revdate)
{
  classname  = new_classname;
  maintainer = new_maintainer;
  revision   = new_revision;
  revdate    = new_revdate;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////// Tcl/Tk Interface ////////////////////////////
//////////////////////////////////////////////////////////////////////////

Tcl_Interp* omfInterp=NULL;  // Used by TkMessage() error handler,
/// some vector file input routines, and the time routines.

BOOL omfInterpNoTk=0;        // True if omfInterp is Tcl-only, i.e.,
                             // setup by Omf_InitLight();

OmfTclPatchLevelStruct Omf_RunningTclPatchLevel={0,0,0,0};

void Omf_SetPatchLevel(Tcl_Interp *interp,OmfTclPatchLevelStruct &tpl)
{
  const char* cmdstr="regsub -all -- {[.abp]} [info patchlevel] "
    "{:&:} junk ; set junk";
  char *cmdbuf=new char[strlen(cmdstr)+1];
  strcpy(cmdbuf,cmdstr);
  Tcl_Eval(interp,cmdbuf);
  delete[] cmdbuf;

  // interp->result holds a string that looks, e.g., like "8:.:0:b:1"
  // The latter two fields may be missing.
#if DEBUGLVL>9
  fprintf(stderr,"PATCHLEVEL STRING: %s\n",interp->result);
#endif
  tpl.major=tpl.minor=tpl.level=tpl.serial=0;

  char *cptr;
  cptr=strtok(interp->result,":.");     // Major
  if(cptr==NULL) goto DONE;
  tpl.major=atoi(cptr);
#if DEBUGLVL>9
  fprintf(stderr,"MAJOR: cptr=%s\n",cptr);
#endif

  cptr=strtok(NULL,":.");               // Minor
  if(cptr==NULL) goto DONE;
  tpl.minor=atoi(cptr);
#if DEBUGLVL>9
  fprintf(stderr,"MINOR: cptr=%s\n",cptr);
#endif
  
  cptr=strtok(NULL,":.");               // Level
  if(cptr==NULL) {
    tpl.level=2;  // Equivalent to 'p'
    tpl.serial=0;
    goto DONE;
  }
  switch (*cptr) {
  case 'a':    tpl.level=0; break;
  case 'b':    tpl.level=1; break;
  default:     tpl.level=2; break;
  }
#if DEBUGLVL>9
  fprintf(stderr,"LEVEL: cptr=%s\n",cptr);
#endif
  
  cptr=strtok(NULL,":.");               // Serial
  if(cptr==NULL) goto DONE;
  tpl.serial=atoi(cptr);
#if DEBUGLVL>9
  fprintf(stderr,"SERIAL: cptr=%s\n",cptr);
#endif

DONE:
  // One could put error checks here.
  return;
}

int Omf_ComparePatchLevel(const OmfTclPatchLevelStruct &tpl,
			  int major,int minor,int level,int serial)
{ // Returns 0 on equal, -1 if tpl<major.minor.level.serial,
  //                  or +1 if tpl>major.minor.level.serial
  if(tpl.major<major)   return -1;  if(tpl.major>major)   return  1;
  if(tpl.minor<minor)   return -1;  if(tpl.minor>minor)   return  1;
  if(tpl.level<level)   return -1;  if(tpl.level>level)   return  1;
  if(tpl.serial<serial) return -1;  if(tpl.serial>serial) return  1;
  return 0;
}

// C-pointer <--> ascii string conversion class.
//   Currently this routine maps things through %lx, but this
// may change in the future to use a hash table.  (I'm assuming
// that void * can be converted to an unsigned long without
// loss of data.  There is a check on this in PtrToAscii for safety.)

const ClassDoc Omf_AsciiPtr::class_doc("Omf_AsciiPtr",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","19-Sep-1997");

const size_t Omf_AsciiPtr::ascii_string_width=2*sizeof(unsigned long)+8;
/// Each byte goes to 2 hex digits (hence 2*), +1 for the trailing
/// null, +2 in case a "0x" is prepended, +5 for safety.

void Omf_AsciiPtr::PtrToAscii(const void* ptr,char* buf)
{
#define MEMBERNAME "PtrToAscii(const void* ptr,char* buf)"
  sprintf(buf,"%lx",(unsigned long)(ptr));
  if(strlen(buf)+1>GetAsciiSize()) {
    FatalError(-1,STDDOC,"Buffer overflow; "
	       "Omf_AsciiPtr::ascii_string_width too small");
  }

  // Check readback
  if(AsciiToPtr(buf)!=ptr) {
    FatalError(-1,STDDOC,"Ptr -> %%lx -> ptr conversion error");
  }

  return;
#undef MEMBERNAME
}

void* Omf_AsciiPtr::AsciiToPtr(const char* buf)
{ // NOTE: ONLY strings produced by Omf_AsciiPtr::PtrToAscii
  //  should ever be sent to this routine.
#define MEMBERNAME "AsciiToPtr(const char* buf,void* &ptr)"
  unsigned long temp;
  sscanf(buf,"%lx",&temp);
  return (void *)temp;
#undef MEMBERNAME
}

// Lightweight Tcl-only initialization for small apps not using
// Tk or event loop
void Omf_InitLight(char *argv0)
{
  if(omfInterp==NULL) {
    Tcl_FindExecutable(argv0);
    omfInterp=Tcl_CreateInterp();
    omfInterpNoTk=1;
    // Initialize the interpreter -- in particular, activate the
    // auto-load library functions of Tcl
    if (Tcl_Init(omfInterp) == TCL_ERROR) {
       panic("Tcl initialization error:\n\n%s\n\n\
As a workaround, set the environment variable\n\
TCL_LIBRARY to the name of a directory\n\
containing an error-free init.tcl.\n\n\
For more information see the INSTALL file.", omfInterp->result);
    }
    // Add directories to auto-loader search path
    // NOTE: Since mmdisp0.0 libraries are being added here by name,
    // this routine will break when we move to a new version of mmdisp.
    Tcl_GlobalEval(omfInterp,"lappend auto_path \
        [file join [file dirname [file dirname [file join \
        [pwd] [info nameofexecutable]]]] library mmdisp0.0]");
    Tcl_GlobalEval(omfInterp,"lappend auto_path \
        [file join [file dirname [file dirname [file join \
        [pwd] [info nameofexecutable]]]] lib mmdisp0.0]");
    atexit(Omf_ExitLight);
  }
}

void Omf_ExitLight()
{
  if(omfInterp!=NULL) {
    Tcl_DeleteInterp(omfInterp);
    omfInterp=NULL;
    omfInterpNoTk=0;
  }
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////// UTILITY ROUTINES ////////////////////////////
//////////////////////////////////////////////////////////////////////////
static char UtilityBuf[BigBufSize];

// Version of atof that allows 'D' or 'd' to be used to denote exponent
double Omf_atof(const char *nptr)
{
  char *cptr;
  if((cptr=strchr(nptr,'d'))==NULL &&
      (cptr=strchr(nptr,'D'))==NULL  )
    return atof(nptr);  // Usual, C-compatible string
  /// The above checks seem to add about 10% overhead for short
  /// strings on a 486/66dx2, and it is faster there to check for
  /// each of 'd' and 'D' separately than to call strcspn(nptr,"dD");
  /// The base atof time on this machine is about 85 microseconds
  /// to start with (compared to about 6 microseconds for atoi).

  // Otherwise, assume we are working with a FORTRAN-style
  // "double precision" string.
  //   On the aforementioned 486, with gcc 2.7.2, this code costs
  // on average about an additional 10% over the above, if statbuf
  // is used.  Calls to new[] and delete[] add another 20%.  NOTE:
  // the strncpy function is really slow on this machine (it takes
  // as long as the atof call itself), so instead of using strncpy
  // to ensure we don't write past the end of statbuf, we take a
  // more flexible approach: see how big the string is, and if it
  // is too big (which is unlikely, but possible), then allocate
  // temporary storage just large enough to hold the string.  The
  // latter is a performance hit, so if it happens frequently one
  // should enlarge the size of statbuf.
  int exp_off=cptr-nptr;
  static char statbuf[256]; // I don't expect many numeric strings this long
  char *buf=statbuf;
  size_t num_len=strlen(nptr)+1;
  if(num_len>sizeof(statbuf)) buf=new char[num_len];

  strcpy(buf,nptr);
  buf[exp_off]='e';         // Replace 'd' or 'D' with 'e'
  double result=atof(buf);

  if(buf!=statbuf) delete[] buf;
  return result;
}

// Prior to Tcl 7.5p1, Tcl_Alloc and Tcl_Free didn't exist, and
// one used malloc & free instead.  omf_Tcl_Alloc and omf_Tcl_Free
// provide backwards compatible substitutes.
char *omf_Tcl_Alloc(int size)
{
#if (TCL_MAJOR_VERSION >= 8) ||                              \
    (TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION  >= 6 ) ||  \
    (TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION  == 5  &&   \
     TCL_RELEASE_LEVEL == 2 && TCL_RELEASE_SERIAL >= 1)
  if(Omf_ComparePatchLevel(Omf_RunningTclPatchLevel,7,5,2,1)>=0) {
    // We're running Tcl >= 7.5p1
    return Tcl_Alloc(size);
  }
#endif
  return (char *)malloc(size);
}

void omf_Tcl_Free(char *ptr)
{
#if (TCL_MAJOR_VERSION >= 8) ||                              \
    (TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION  >= 6 ) ||  \
    (TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION  == 5  &&   \
     TCL_RELEASE_LEVEL == 2 && TCL_RELEASE_SERIAL >= 1)
  if(Omf_ComparePatchLevel(Omf_RunningTclPatchLevel,7,5,2,1)>=0) {
    // We're running Tcl >= 7.5p1
    Tcl_Free(ptr);
    return;
  }
#endif
  free(ptr);
  return;
}

long Omf_GetClicks()
{ // Accesses the Tcl "clock clicks" command
  const char* cmdstr="clock clicks";
  char *cmdbuf=new char[strlen(cmdstr)+1];
  strcpy(cmdbuf,cmdstr);
  Tcl_Eval(omfInterp,cmdbuf);
  delete[] cmdbuf;

  // interp->result holds a string containing a "high resolution
  // system-dependent integer time value."

  return atol(omfInterp->result);
}

void Omf_TclUpdate(int idletasks)
{ // Calls the Tcl/Tk command 'update'.
  // If import idletasks is !=0, then calls 'update idletasks'
  if(omfInterp==NULL) return;
  const char* cmdstr="update";
  if(idletasks) cmdstr="update idletasks";
  char *cmdbuf=new char[strlen(cmdstr)+1];
  strcpy(cmdbuf,cmdstr);
  Tcl_Eval(omfInterp,cmdbuf);
  delete[] cmdbuf;
}

void Omf_TclNativeFilename(String &filename)
{
  // Calls Tcl command 'file nativename' to convert "filename" from Tcl
  // standard format (with /'s) to a platform specific form.  If the
  // call is successful, and the resulting string is smaller than
  // bufsize, then overwrites filename with the returned result.
  // Otherwise filename is left unchanged.
  if(omfInterp==NULL) return;
  const char* cmdfmt="file nativename [file join %s]";
  if(strlen(cmdfmt)+filename.Length()+2>sizeof(UtilityBuf)) return;
  sprintf(UtilityBuf,cmdfmt,filename.GetStr());
  if(Tcl_Eval(omfInterp,UtilityBuf)==TCL_OK) {
    // Note: If an older version of Tcl is running that does not
    // support the "file nativename" command, then Tcl_Eval will
    // always return TCL_ERROR.
    filename=omfInterp->result;
  }
}

// A slightly portable version of fopen
FILE *Omf_FOpen(const char *path,const char *mode)
{
  String filename(path);
  Omf_TclNativeFilename(filename);
  return fopen(filename.GetStr(),mode);
}


//////////////////////////////////////////////////////////////////////////
///////////////////////////// ERROR HANDLERS /////////////////////////////
//////////////////////////////////////////////////////////////////////////

static char ErrorBuf[BigBufSize];

// Things worth dying over...
void FatalError(INT4m errorcode,
		const CHAR *membername,
		const ClassDoc &class_doc,
		const CHAR *filename,INT4m lineno,
		const CHAR *fmt,...)
{
  va_list arg_ptr;
  char *cptr;
  sprintf(ErrorBuf,
	  "Fatal error occurred in %s::%s (File %s, line %d)\n"
	  "Class revision/date: %s/%s\n"
	  "Class maintainer: %s\nERROR: ",
	  class_doc.classname,membername,filename,lineno,
	  class_doc.revision,class_doc.revdate,
	  class_doc.maintainer);
  cptr=strchr(ErrorBuf,'\0'); // Note: Return type of sprintf is OS dependent
  va_start(arg_ptr,fmt);
  vsprintf(cptr,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_ERROR,ErrorBuf,errorcode);

  // We shouldn't get to here, but just in case...
  exit(1);
}

// Fatal Error handler for non-class functions
void PlainError(INT4m errorcode,const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_ERROR,ErrorBuf,errorcode);

  // We shouldn't get to here, but just in case...
  exit(1);
}


// Warnings
void NonFatalError(const CHAR *membername,
		   const ClassDoc &class_doc,
		   const CHAR *filename,INT4m lineno,
		   const CHAR *fmt,...)
{
  va_list arg_ptr;
  char *cptr;
  sprintf(ErrorBuf,
	  "Nonfatal error occurred in %s::%s (File %s, line %d)\n"
	  "Class revision/date: %s/%s\n"
	  "Class maintainer: %s\nWARNING: ",
	  class_doc.classname,membername,filename,lineno,
	  class_doc.revision,class_doc.revdate,
	  class_doc.maintainer);
  cptr=strchr(ErrorBuf,'\0'); // Note: Return type of sprintf is OS dependent
  va_start(arg_ptr,fmt);
  vsprintf(cptr,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_WARNING,ErrorBuf,0);
}

// Nonfatal error handler for non-class functions
void PlainWarning(const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_WARNING,ErrorBuf,0);
}

// Error & warning messages storage
String MessageLocker::msg;

int MessageLocker::Append(const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);
  msg.Append(ErrorBuf,sizeof(ErrorBuf));
  return 0;
}

int MessageLocker::GetMessage( String& _msg)
{
  _msg=msg;  // Duplicates
  return 0;
}

// Informational messages
int Message(const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  return TkMessage(MSG_INFO,ErrorBuf,0);
}

// C-interface to Tcl/Tk error boxes.
INT4m TkMessage(MessageType mt,const CHAR *msg,INT4m exitcode)
{
  static char buf[BigBufSize];

  switch(mt) {
  case MSG_INFO:
    sprintf(buf,"OMF_InfoDialog {%s} {mmDisp Info} info",msg);
    break;
  case MSG_WARNING:
    sprintf(buf,"OMF_NonFatalError {%s} {} %d",msg,exitcode); break;
  case MSG_ERROR:
    sprintf(buf,"OMF_FatalError {%s} {} %d",msg,exitcode);    break;
  default:
    sprintf(buf,"OMF_FatalError {Unknown MessageType (%d) detected"
	    " in TkMessage(); Associated message: %s} {} 1",
	    int(mt),msg);
    break;
  }

/**/ fprintf(stderr,"Inside TkMessage; omfInterp=%p, omfInterpNoTk=%d\n",
/**/         omfInterp,omfInterpNoTk);
  if(omfInterp==NULL || omfInterpNoTk) {
    // Tcl interp not setup, or setup for Tcl-only (no Tk)
    // Write to MessageBuffer
    strcat(buf,"\n");
    MessageLocker::Append(buf);
    if(mt==MSG_ERROR) {
      // If fatal error, dump MessageLocker to stderr and exit
      String amsg;
      MessageLocker::GetMessage(amsg);
      fprintf(stderr,"%s\nFATAL ERROR\n",amsg.GetStr());
      fflush(stderr);
      exit(exitcode);
    }
    return -1;
  }

  Tcl_GlobalEval(omfInterp,buf);  // We could check for errors
/**/ fprintf(stderr,"Interp result: ->%s<-\n",omfInterp->result);
  // from Tcl_GlobalEval, but to where would we send a message???
  return atoi(omfInterp->result);
}


//////////////////////////////////////////////////////////////////////////
// Greatest common divisor via Euclid's algorithm
int gcd(int m,int n)
{
  int temp;
  while((temp=(m%n))>0) { m=n; n=temp; }
  return n;
}


//////////////////////////////////////////////////////////////////////////
// Random number generator

double UnifRand(unsigned int seed)
{ // Returns random value in [0,1] with unif. distrib.
  // seed>0 reinitializes the generator with "seed"
  
  static int first=1;

  if(seed>0 || first) {
    int exercise=100;
    if(seed==0) { // Use random seed
      seed=Omf_GetClicks();
    }
    OMF_SRANDOM(seed);
    for(first=0;first<exercise;first++)
      OMF_RANDOM();  // Exercise generator
    first=0;
  }
  return ((double)OMF_RANDOM())/((double)OMF_RANDOM_MAX);
}

//////////////////////////////////////////////////////////////////////////
// Portable case insensitive string comparison, modeled after
// strcasecmp().
int Omf_StrCaseCmp(const char *s1, const char *s2)
{
  const char *cptr1=s1,*cptr2=s2;
  int c1,c2;
  do {
    c1=tolower(*(cptr1++));
    c2=tolower(*(cptr2++));
  } while(c1==c2 && c1!='\0');
  if(c1<c2) return -1;
  if(c1>c2) return  1;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Routine to kill zombie children under unix.  This is a just NOP
// returning TCL_OK under Windows.
int CullZombies(ClientData,Tcl_Interp*,int,char**)
{
#if (SYSTEM_TYPE == UNIX)
  while(waitpid(-1,NULL,WNOHANG)>0); // Kill any zombie'd children
#endif
  return TCL_OK;
}


//////////////////////////////////////////////////////////////////////////
// Byte reversal routines;  Converts between big-endian
// and little-endian representations.

REAL8 flip(REAL8 input)
{ // This can be used on any 8 byte wide type;  Probably too
  // big to be inlined.
  UCHAR temp;
  union { UCHAR byte[8]; REAL8 r; } word;
  word.r=input;
  temp=word.byte[3]; word.byte[3]=word.byte[4]; word.byte[4]=temp;
  temp=word.byte[2]; word.byte[2]=word.byte[5]; word.byte[5]=temp;
  temp=word.byte[1]; word.byte[1]=word.byte[6]; word.byte[6]=temp;
  temp=word.byte[0]; word.byte[0]=word.byte[7]; word.byte[7]=temp;
  return word.r;
}

void swap_bytes(void *buf,INT4m buflen)
{
  swab((char *)buf,(char *)buf,buflen);
}

void swap_words4(void *buf,INT4m count)
{
  UINT4 *bufptr;
  bufptr=(UINT4 *)buf;
  for(;count>0;++bufptr,--count) (*bufptr)=flip(*bufptr);  
}

void swap_words8(void *buf,INT4m count)
{
  REAL8 *bufptr;  // Type is irrelevant, as long as it is 8 bytes wide
  bufptr=(REAL8 *)buf;
  for(;count>0;++bufptr,--count) (*bufptr)=flip(*bufptr);  
}

void swap_buf(void *buf,INT4m wordsize,INT4m wordcount)
{
  // Converts the array 'buf' from big endian to little endian
  // (or vice versa), where each word is of width 'wordsize',
  // and there are 'wordcount' elements (words) in the array.
  if(wordsize<2) return;  // Nothing to do
  unsigned char *ptr0,*ptr1,*ptr2;
  for(ptr0=(unsigned char *)buf;wordcount>0;wordcount--,ptr0+=wordsize) {
      ptr1=ptr0;
      ptr2=ptr0+wordsize-1;
      for(int j=0;;) {
	unsigned char ctemp;
	ctemp= *ptr1;  *ptr1 = *ptr2; *ptr2 = ctemp;
	if(++j >= wordsize/2) break;
	++ptr1; --ptr2;
      }
    }
}

//////////////////////////////////////////////////////////////////////////
// String class
//
const ClassDoc String::class_doc("String",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","18-Mar-1997");

void String::Empty()
{
#define MEMBERNAME "Empty"
  if(str!=NULL) delete[] str;
  str=new char[1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  str[0]='\0';
#undef MEMBERNAME
}

String::String()
{
#define MEMBERNAME "String"
  str=NULL;
  Empty();
#undef MEMBERNAME
}

String::String(const char *cptr)
{
#define MEMBERNAME "String(const char *cptr)"
  if(cptr==NULL) str=new char[1];
  else str=new char[strlen(cptr)+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  if(cptr==NULL) str[0]='\0';
  else strcpy(str,cptr);
#undef MEMBERNAME
}

String::String(const String &rhs)
{
#define MEMBERNAME "String(const String &)"
  str=new char[rhs.Length()+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,rhs.str);
#undef MEMBERNAME
}

String::~String()
{
#define MEMBERNAME "~String"
  delete[] str;
#undef MEMBERNAME
}

String& String::operator=(const String &rhs)
{ // Assignment operator
#define MEMBERNAME "operator=(const String &)"
  if(&rhs==this) return *this;  // Nothing to do!
  delete[] str;
  str=new char[rhs.Length()+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,rhs.str);
  return *this;
#undef MEMBERNAME
}

String& String::operator=(const char *cptr)
{ // Assignment operator
#define MEMBERNAME "operator=(const char *)"
  if(cptr==NULL) FatalError(-1,STDDOC,"Null pointer passed (cptr)");
  if(cptr==str) return *this;  // Nothing to do!
  delete[] str;
  str=new char[strlen(cptr)+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,cptr);
  return *this;
#undef MEMBERNAME
}

INT4m StringCompare(const String &a,const String &b)
{ // Replicates strcmp
  return strcmp(a.str,b.str);
}

INT4m StringSpan(const String &a,const char *b)
{ // Replicates strspn
  return strspn(a.str,b);
}

// Append functions; grows str as needed.  The char version appends
// up to the first '\0' character or maxlen, which ever happens first.
// Both return *this.
String& String::Append(const String& appendstr)
{
  if(appendstr.str[0]=='\0') return *this; // Nothing to do.
  int len1=Length();  int len2=appendstr.Length();
  int len3=len1+len2;
  char *newstr=new char[len3+1];
  memcpy(newstr,str,len1);
  memcpy(newstr+len1,appendstr,len2);
  newstr[len3]='\0';
  delete[] str;
  str=newstr;
  return *this;
}

String& String::Append(const char *appendarr,int maxlen)
{
  // Determine length of string to append
  int len2=0;
  while(len2<maxlen && appendarr[len2]!='\0') len2++;
  if(len2==0) return *this; // Nothing to do.
  int len1=Length();
  int len3=len1+len2;
  char *newstr=new char[len3+1];
  memcpy(newstr,str,len1);
  memcpy(newstr+len1,appendarr,len2);
  newstr[len3]='\0';
  delete[] str;
  str=newstr;
  return *this;
}

// Function to trim leading & trailing whitespace (as determined by
// 'isspace' function) from string.
// NOTE: This routine currently operates "in-place"
String& String::Trim()
{
  char *cptr1,*cptr2,*cptr3;
  cptr1=str;
  while(*cptr1!='\0' && isspace(*cptr1)) cptr1++;
  if(*cptr1=='\0') {
    str[0]='\0'; // Blank string
    return *this;
  }
  // From this point down, we know *cptr1 is non-space
  cptr2=strchr(cptr1,'\0')-1;
  while(isspace(*cptr2)) cptr2--;
  *(++cptr2)='\0'; // cptr2 points to first whitespace character in trailer

  if(cptr1!=str) {
    // Down copy string between cptr1 & cptr2
    cptr3=str;
    while(cptr1<cptr2) {
      *(cptr3++) = *(cptr1++);
    }
  }
  return *this;
}

// Function to collapse runs of whitespace into single blanks.
// Also removes leading and trailing whitespace.  Both normal
// function and member function versions.
void CollapseStr(char *str)
{
  int i1,i2,lastblank;  char ch;
  for(i1=i2=0,lastblank=1;str[i2]!='\0';i2++) {
    if(isspace(ch=str[i2])) {
      if(!lastblank) { str[i1++]=' '; lastblank=1; }
    }
    else { str[i1++]=ch; lastblank=0; }
  }
  if(i1>0 && str[i1-1]==' ') str[i1-1]='\0';  // Remove trailing whitespace
  else                       str[i1]='\0';
}

String& String::CollapseStr()
{
  ::CollapseStr(str);
  return *this;
}

// Converts string to lowercase, using C library function tolower().
// Both normal function and member function versions.
void LowerString(char *str)
{
  char *cptr=str;
  while(*cptr!='\0') {
    *cptr=tolower(*cptr);
    cptr++;
  }
}

String& String::ToLower()
{
  LowerString(str);
  return *this;
}


