/* FILE: misc.h                 -*-Mode: c++-*-
 *
 * Useful miscellaneous stuff.
 *
 */

#ifndef MISC_H
#define MISC_H

#include "machine.h"
#include "oommf.h"

#include <math.h>
#include <stdarg.h>
#include <tcl.h>

/* End includes */     /* Optional directive to build.tcl */

#ifndef PI
#define PI             3.14159265358979323846
#endif

#ifndef SQRT2
#define SQRT2          1.41421356237309504880
#endif

#define SCRATCH_BUF_SIZE 1024  // Size of scratch space buffers
#define ASCIIPTR_BUF_SIZE (POINTERASCIIWIDTH+16)
/// Room + extra for %p representation of pointers


extern int Verbosity;
#ifndef DEBUGLVL
#define DEBUGLVL 100  // 0 => No debug code, 100 => Full debug code
#endif 

// Several useful macros
#define SQ(x) ((x)*(x))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))
#define ROUND(x) ((int)rint(x))

int gcd(int m,int n);  /* Greatest Common Divisor */

//////////////////////////////////////////////////////////////////////////
// Class documentation structure
struct ClassDoc {
public:
  const CHAR *classname;
  const CHAR *maintainer;
  const CHAR *revision;
  const CHAR *revdate;
  ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
	   const CHAR *new_revision,const CHAR *new_revdate);
};

//////////////////////////////////////////////////////////////////////////
// String class
class String
{
  friend INT4m StringCompare(const String &,
			     const String &); // Replicates strcmp
  friend INT4m StringSpan(const String &,
			  const char *); // Replicates strspn
private:
  static const ClassDoc class_doc;
  char *str;
  void Empty(void);
public:
  String(void);                      // Default constructor
  String(const String &copyobj);     // Copy constructor
  String(const char *cptr);
  String& operator=(const String &rhs);  // Duplicates string
  String& operator=(const char *cptr);   // Ditto
  ~String();

  const char* GetStr(void) { return (const char *)str; }

  char& operator[](int offset) { return str[offset]; }
  // The following [] operator can be used on 'const String' objects.
  const char& operator[](int offset) const { return str[offset]; }

  operator const char*() const { return (const char *)str; }
  // Note: In the last 3 functions, the returned pointer may become
  //  invalid if later any of the non-const String member functions
  //  are called.

  // Length *not* including trailing null character.
  int Length(void) const { return strlen(str); }

  // Append functions; grows str as needed.  The char version appends
  // up to the first '\0' character or maxlen, which ever happens first.
  // Either way the last char in this->str is set to '\0'.
  // Both functions return *this.
  String& Append(const String& appendstr);
  String& Append(const char *appendarr,int maxlen);

  // Function to trim leading & trailing whitespace (as determined by
  // 'isspace' function) from string.
  String& Trim();

  // Function to collapse runs of whitespace into single blanks.
  // Also removes leading and trailing whitespace.
  String& String::CollapseStr();

  // Interface to C tolower function
  String& ToLower();

  void Free(void) { Empty(); }
};

// Normal C char[] functions.  Use by above String class, but
// may also be called from outside for operations on non-String
// char arrays.
void LowerString(char *str);
void CollapseStr(char *str);

//////////////////////////////////////////////////////////////////////////
// Tcl/Tk Interface
extern Tcl_Interp *omfInterp;  // Tcl interpreter
extern BOOL omfInterpNoTk;     // True if omfInterp is Tcl-only

EXTERN void panic _ANSI_ARGS_(TCL_VARARGS(char *,format));
#if (SYSTEM_TYPE==WINDOWS)
EXTERN void Omf_WinPanic _ANSI_ARGS_(TCL_VARARGS(char *,format));
#endif

typedef struct {
  int major;        // Tcl major version
  int minor;        //     minor version
  int level;        //     release level
  int serial;       //     release serial
} OmfTclPatchLevelStruct;

// Patch level at run time (which may differ from compile time pl.)
extern OmfTclPatchLevelStruct Omf_RunningTclPatchLevel;

void Omf_SetPatchLevel(Tcl_Interp *interp,OmfTclPatchLevelStruct &tpl);
int Omf_ComparePatchLevel(const OmfTclPatchLevelStruct &tpl,
			  int major,int minor,int level,int serial);
/// Returns 0 on equal, -1 if tpl<major.minor.level.serial,
///                  or +1 if tpl>major.minor.level.serial

// C-pointer <--> ascii string conversion class.
//   Currently this routine maps things through %x or %lx, but this
// may change in the future to use a hash table.  The client should
// first call GetAsciiSize to determine how large a buffer is needed
// to hold the ascii-fied pointer string, then call PtrToAscii to
// get a string representation of the ptr.  This can be passed across
// to Tcl procs as needed.  Later, when a C routine is called by a
// Tcl proc having such a string, this string can be converted back
// to a valid ptr via AsciiToPtr.  NOTE: AsciiToPtr should ONLY be
// called with strings produced by PtrToAscii.  There is currently
// no check on this, but there will be if this class gets modified
// to use a hash table lookup scheme.
//
class Omf_AsciiPtr
{
private:
  static const ClassDoc class_doc;
  static const size_t ascii_string_width;
public:
  static size_t GetAsciiSize() { return ascii_string_width; }

  static void PtrToAscii(const void* ptr,char* buf);
  /// It is the responsibility of the calling routine to insure
  /// that buf is dimensioned to at least GetStringSize() bytes.

  static void* AsciiToPtr(const char* buf);
  /// NOTE 1: ONLY strings produced by Omf_AsciiPtr::PtrToAscii
  ///  should ever be sent to this routine. 
  /// Note 2: For consistency, I would prefer to make this function
  ///    static void AsciiToPtr(const char* buf,void * &ptr)
  ///  but gcc 2.7.2.1 (at least) screws up implicit casting
  ///  to void *&, apparently producing a temporary which never
  ///  gets written back into ptr.  So the above assignment seems
  ///  a safer option.  (The alternative is to be _real_ sure
  ///  that the client routine does an explicit cast to void *&.)
};

void Omf_InitLight(char *argv0);
extern "C" void Omf_ExitLight();

//////////////////////////////////////////////////////////////////////////
// Utility routines
double Omf_atof(const char *nptr); // Version of atof that allows 'D' or
                                   // 'd' to be used to denote exponent.

char *omf_Tcl_Alloc(int size);  // Provides backwards compatibility
void omf_Tcl_Free(char *ptr);   // for older versions of Tcl.

// Timing routines
long Omf_GetClicks();  // Interface to Tcl "clock clicks" call.

// C interface to the Tcl/Tk update command
void Omf_TclUpdate(int idletasks=1);

// C interface to the Tcl 'file nativename' command.
void Omf_TclNativeFilename(String &filename);

// A version of fopen that calls Omf_TclNativeFilename().
FILE *Omf_FOpen(const char *path,const char *mode);

// Random number generator
double UnifRand(unsigned int seed=0);
/// Returns random value in [0,1] with unif. distrib.
/// seed>0 => reinitialize generator with "seed"
/// Routines needing random numbers should call this rather than
/// the system calls rand() or random(), which may be system dependent.

// Portable case insensitive string comparison, modeled after
// strcasecmp().
int Omf_StrCaseCmp(const char *s1, const char *s2);

// Routine to kill zombies under unix.  This is a NOP under windows
TCLCMD CullZombies;

//////////////////////////////////////////////////////////////////////////
// Error codes
extern const char* ErrArrOutOfBounds;
extern const char* ErrBadFile;
extern const char* ErrBadFileHeader;
extern const char* ErrBadFileOpen;
extern const char* ErrInputEOF;
extern const char* ErrOutputEOF;
extern const char* ErrBadParam;
extern const char* ErrDispInit;
extern const char* ErrMisc;
extern const char* ErrNoMem;
extern const char* ErrOverflow;
extern const char* ErrProgramming;

//////////////////////////////////////////////////////////////////////////
// Error handlers
// Note: Use preprocessor macros __FILE__ and __LINE__ to get
//       'filename' and 'lineno' respectively.
void FatalError(INT4m errorcode,
		const CHAR *membername,
		const ClassDoc &class_doc,
		const CHAR *filename,INT4m lineno,
		const CHAR *fmt,...);
// Doesn't return; exits with code 'errorcode'

void NonFatalError(const CHAR *membername,
		   const ClassDoc &class_doc,
		   const CHAR *filename,INT4m lineno,
		   const CHAR *fmt,...);
// Prints warning message and returns.

// Versions of the above for non-class functions
void PlainWarning(const CHAR *fmt,...);
void PlainError(INT4m errorcode,const CHAR *fmt,...);

// Class to store error & warning messages in a global String
class MessageLocker
{
private:
  static String msg;
public:
  static int Reset() { msg.Free(); return 0; }
  static int Append(const CHAR *fmt,...);
  static int GetMessage( String& _msg);   // Dup's msg into _msg.
};

// Routine to print message to standard (user) output channel
int Message(const CHAR *message_fmt,...);

// C-interface to Tcl/Tk error boxes.  The above routines map
// into this.  This routine is defined in tkcmds.cc
enum MessageType { MSG_INFO, MSG_WARNING, MSG_ERROR };
INT4m TkMessage(MessageType mt,const CHAR *msg,INT4m exitcode=1);

// If you #define MEMBERNAME at the start of each member definition,
// then you can use the following macro to expand to 4 of the fields
// needed by the above error handlers.
#define STDDOC MEMBERNAME,class_doc,__FILE__,__LINE__

//////////////////////////////////////////////////////////////////////////
// Byte reversal routines, for conversion between big-endian
// and little-endian representations
inline UINT2 flip(UINT2 input);
inline INT2  flip(INT2  input);
inline INT4  flip(INT4  input);
inline REAL4 flip(REAL4 input);
REAL8 flip(REAL8 input);  // Probably too big to be inlined.
void swap_bytes(void *buf,INT4m buflen);
void swap_words4(void *buf,INT4m count);
void swap_words8(void *buf,INT4m count);
void swap_buf(void *buf,INT4m wordsize,INT4m wordcount);
  // Converts the array 'buf' from big endian to little endian
  // (or vice versa), where each word is of width 'wordsize',
  // and there are 'wordcount' elements (words) in the array.

// inline codes...
inline UINT2 flip(UINT2 input)
{
  UCHAR temp;
  union { UCHAR byte[2]; UINT2 i; } word;
  word.i=input;
  temp=word.byte[0]; word.byte[0]=word.byte[1]; word.byte[1]=temp;
  return word.i;
}

inline INT2 flip(INT2 input)
{
  union { UINT2 u; INT2 i; } word;
  word.i=input;
  word.u=flip(word.u);  // signed <-> unsigned casting is
  return word.i;        // not well specified in C/C++
}

inline UINT4 flip(UINT4 input)
{
  UCHAR temp;
  union { UCHAR byte[4]; UINT4 i; } word;
  word.i=input;
  temp=word.byte[1]; word.byte[1]=word.byte[2]; word.byte[2]=temp;
  temp=word.byte[0]; word.byte[0]=word.byte[3]; word.byte[3]=temp;
  return word.i;
}

inline INT4 flip(INT4 input)
{
  union { UINT4 u; INT4 i; } word;
  word.i=input;  
  word.u=flip(word.u);  // signed <-> unsigned casting is
  return word.i;        // not well specified in C/C++
}

inline REAL4 flip(REAL4 input)
{
  union { UINT4 u; REAL4 r; } word;
  word.r=input;  
  word.u=flip(word.u);
  return word.r;
}

//////////////////////////////////////////////////////////////////////////
// Routines to interchange two values
template<class T> inline void swap(T &x,T &y) {
  T temp=x; x=y; y=temp;
}

////////////////////////////////////////////////////////////////////
// Temporary porting aids

#define ErrBadFileOpenS ""

#define StandardInput  "Read input from stdin"
#define StandardOutput "Write output to stdout"

class StopWatch {
private:
public:
  StopWatch(void) {}
  void Start(void) {}
  void Stop(void) {}
  void Print(...) {}
};


#endif // MISC_H

